#include "palclass.cpp"
#include "geoclass.cpp"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <stdio.h>
#include <dos.h>

/**************************** declarations *****************************/
void spill(int spillx, int spilly, int colour);
unsigned char readkbd(void);
void erase(int x1, int y1, int x2, int y2, int xoffset, int yoffset, unsigned char far *pic);
char touching(int x1, int y1, char *pic1, int x2, int y2, char *pic2, char mode);
void cls(void);
char load_font(unsigned char *name);
void unload_font(void);
void save_font(unsigned char *name);
void drawtext(int x, int y, unsigned char *string, unsigned char colour, char mode);
void loadpic(char far *image, int arraysize, unsigned char *filename);
void savepic(char far *image, int arraysize, unsigned char *filename);
void circle(int cx, int cy, int radius, unsigned char colour, char mode);
void put(int x1, int y1, unsigned char far *picture, char mode);
void get(int x1, int y1, int x2, int y2, unsigned char far *picture);
void box(int x1, int y1, int x2, int y2, int colour, char mode);
void line(int x1, int y1, int x2, int y2, unsigned char colour, char mode);
char getpixel(int x, int y);
char sgn(int val);
void putpixel(int x, int y, unsigned char colour, char mode);
void locate(char x, char y);
void readmousepos(int *mx, int *my);
int mousex(void);
int mousey(void);
void showmouse(int onoff);
void initmouse(void);
char buttonpressed(void);
void initscrn(char mode);
void readmousediff(void);
char blackcharstatus(void);
void setmousepos(int x, int y);
void setmouselimits(int x1, int y1, int x2, int y2);
char turnoffmouse(void);
void set_colour(int c, char r, char g, char b);
char kbscan(void);
unsigned int counter(void);
unsigned char *taketext(unsigned char *title, char *buttonused, char numbuttons, ...);
unsigned char choose(unsigned char *title, char numbuttons, ...);
void writebit(FILE *fout, char bitval);
char readbit(FILE *fin);
void rotput(int cx, int cy, int xoffset, int yoffset, float ang, float scale, unsigned char far *pic, char mode);
void roterase(int x1, int y1, int x2, int y2, int cx, int cy, float ang, float scale, unsigned char far * pic);
void scaleput(int x1, int y1, float xscale, float yscale, unsigned char far *pic, char mode);
void calcrotarea(int cx, int cy, int xoffset, int yoffset, float ang, float scale, unsigned char far *pic, int *x1, int *y1, int *x2, int *y2);
void calccorners(int cx, int cy, int xoffset, int yoffset, float ang, float scale, unsigned char far *pic, int *x1, int *y1, int *x2, int *y2, int *x3, int *y3, int *x4, int *y4);
void putfont(int x1, int y1, unsigned char far *picture, unsigned char colour, char mode);
unsigned char csmode(void);
void set_keyboard_rate(unsigned char rate, unsigned char delay);
void drawpoly(int *x, int *y, unsigned char colour, char mode);
void hline(int x1, int x2, int y, unsigned char colour, char mode);
unsigned char getch(void);
unsigned char kbhit(void);
void set_count_rate(int rate);
void drawchar(unsigned char c);
void adjust_value(char *title, int *value, int min, int max);

void handle_keypress(void);
void handle_option(int mop, int copt);
void resize_grid(void);
void load_picture(void);
void save_picture(char *filename);
void clear_screen(void);
int on_palette(int x, int y);
void draw_palette(void);
int picx(int x);
int picy(int y);
void draw_grid(void);
void handle_drawing(int px, int py);
int current_option(int my, int menuy, int menulength);
int on_open_menu(int mx, int my, int menux, int menuy,
				 int menuxsize, int menuysize);
void close_menu(char *undermenu, int menux, int menuy, int *menu_open);
void open_menu(int cm, int *menulength, char *undermenu, int *mnux, int *mnuy);
int current_menu(int mx, int my, int nummenus);
void drawmenus(int nummenus);
void load_menus(unsigned char *filename, int *nummenus, int *menulength);
void update_grid(void);
void update_grid_area(int x1, int y1, int x2, int y2);
void handle_spill(int px, int py);
void handle_pencil(int px, int py);
void handle_line(int px, int py);
void handle_box(int px, int py);
void handle_empty_box(int px, int py);
void handle_airbrush(void);
void handle_paintbrush(void);
void optimize_grid(void);
void handle_replacement(void);
void update_pattern(void);

/**************************** definitions ******************************/
#define tabsize 4
#define pi 3.141592653589793238462643383279502884197169399375

/************************** Global Variables ***************************/
unsigned char far *vmem = (unsigned char far *)0xA0000000;
vectortype<int> mousediff;
unsigned char **font = NULL, xfontsize, yfontsize;

#define palettex 220
#define palettey 5
#define palettescale 5
#define picturex 220
#define picturey 95
#define maxmenuwidth 20
#define maxmenulength 16
#define maxmenus 4
#define DEFAULT_GRID_COLOUR 8
#define menubordersize 3

enum drawmodetype {
	PENCIL, PAINTBRUSH, AIRBRUSH, CIRCLE_EMPTY, CIRCLE_FULL,
	BOX_EMPTY, BOX_FULL, SPILL, LINE, DITHER, REPLACE, HIGHLIGHT,
	ENVELOP
};

int minmenux, minmenuy, maxmenux, maxmenuy;
int gridx = 5, gridy = 15, gridcolour = DEFAULT_GRID_COLOUR, grid_showing = 1;
int nummenus = 0, xgridsize = 24, ygridsize = 20, scale;
int airbrush_rate = 10, airbrush_range = 10;
unsigned char menu[maxmenus][maxmenulength][maxmenuwidth];
unsigned char colour = 15, textcolour = 65;
char done = 0, picname[257], highlight_direction;
drawmodetype drawmode = PENCIL;
palettetype palette;

void main(void){
	/*** variables, etc. ***/
	unsigned char smode, *undermenu, memgood;
	int menulength[maxmenus], mx, my, menu_open = -1;
	int cm = -1, menux, menuy, px, py, n;

	/*** initialize / load / etc. stuff ***/
	scale = (int)(palettex - gridx - 2) / xgridsize;
	picname[0] = NULL;
	load_font("6X6.BLF");
	memgood = font != NULL;
	undermenu = (unsigned char *)
		malloc((xfontsize * maxmenuwidth + 6) *
			   (yfontsize * maxmenulength + 6) + 2);
	memgood &= (undermenu != NULL);

	if(memgood){
		smode = csmode();
		initscrn(0x13);
		initmouse();
		load_menus("EDPAT.MNU", &nummenus, menulength);
		drawmenus(nummenus);
		draw_grid();
		draw_palette();
		showmouse(1);
		minmenuy = 0;
		maxmenuy = minmenuy + yfontsize + 2 * menubordersize;
		minmenux = maxmenux = 0;
		for(n = 0; n < nummenus; n++)
			maxmenux += xfontsize * (strlen(menu[n][0]));

		/*** main loop ***/
		while(!done){

			mx = mousex();
			my = mousey();
			if(kbhit()) handle_keypress();
			if(buttonpressed()){
				if((menu_open != -1) && on_open_menu(mx, my, menux, menuy, undermenu[0], undermenu[1])){

					int copt = current_option(my, menuy, menulength[menu_open]);
					int mop = menu_open;
					close_menu(undermenu, menux, menuy, &menu_open);
					handle_option(mop, copt);

					while(buttonpressed());
				}else{
					cm = current_menu(mx, my, nummenus);

					if(cm == -1){ /*button pressed off the menu bar*/
						if(menu_open != -1)
							close_menu(undermenu, menux, menuy, &menu_open);
						px = picx(mx);
						py = picy(my);
						if((px != -1) && (py != -1)){
							handle_drawing(px, py);
							while(buttonpressed() && (px == picx(mousex())) && (py == picy(mousey())));
						}else if(on_palette(mx, my)){
							showmouse(0);
							box(palettex + (colour % 16) * palettescale,
								palettey + (int)(colour / 16) * palettescale,
								palettex + (colour % 16 + 1) * palettescale,
								palettey + (int)(colour / 16 + 1) * palettescale, 8, 4);
							colour = (int)((mx - palettex) / palettescale)
								   + 16 * (int)((my - palettey) / palettescale);
							box(palettex + (colour % 16) * palettescale,
								palettey + (int)(colour / 16) * palettescale,
								palettex + (colour % 16 + 1) * palettescale,
								palettey + (int)(colour / 16 + 1) * palettescale, 15, 4);
							showmouse(1);
						}

					}else{ /*button pressed on the menu bar*/
						if(cm != menu_open){
							if(menu_open != -1)
								close_menu(undermenu, menux, menuy, &menu_open);

							showmouse(0);
							open_menu(cm, menulength, undermenu, &menux, &menuy);
							menu_open = cm;
							showmouse(1);
							while(buttonpressed());
						}
					}
				}
			}
		}
		showmouse(0);
		initscrn(smode);
	}
	if(undermenu) free(undermenu);
	unload_font();
}

void handle_keypress(void){
	int x, y;
	switch(getch()){
		case(0):
			showmouse(0);
			box(palettex + (colour % 16) * palettescale,
				palettey + (int)(colour / 16) * palettescale,
				palettex + (colour % 16 + 1) * palettescale,
				palettey + (int)(colour / 16 + 1) * palettescale, 8, 4);
			switch(getch()){
				case 'H':
					colour -= 16;
				break;
				case 'K':
					colour -= 1;
				break;
				case 'M':
					colour = (colour + 1) % 256;
				break;
				case 'P':
					colour = (colour + 16) % 256;
				break;
			}
			box(palettex + (colour % 16) * palettescale,
				palettey + (int)(colour / 16) * palettescale,
				palettex + (colour % 16 + 1) * palettescale,
				palettey + (int)(colour / 16 + 1) * palettescale, 15, 4);
			showmouse(1);
		break;
		case 27:done = 1;break;
		case 9:
			x = mousex();
			y = mousey();
			showmouse(0);
			box(palettex + (colour % 16) * palettescale,
				palettey + (int)(colour / 16) * palettescale,
				palettex + (colour % 16 + 1) * palettescale,
				palettey + (int)(colour / 16 + 1) * palettescale, 8, 4);
			colour = getpixel(picturex + picx(x), picturey + picy(y));
			box(palettex + (colour % 16) * palettescale,
				palettey + (int)(colour / 16) * palettescale,
				palettex + (colour % 16 + 1) * palettescale,
				palettey + (int)(colour / 16 + 1) * palettescale, 15, 4);
			showmouse(1);
		break;
	}
}

void handle_option(int mop, int copt){
	unsigned char c, junk[4][2], palname[256];
	if(!strcmp(menu[mop][copt], "Save as...")){
		showmouse(0);
		strcpy(picname, taketext("What would you like to call it?", &c, 2, "Okay", 13, "Cancel", 27));
		showmouse(1);
		if(picname[0] && !c)
			save_picture(picname);
	}else if(!strcmp(menu[mop][copt], "Save")) save_picture(picname);
	else if(!strcmp(menu[mop][copt], "Resize Grid")) resize_grid();
	else if(!strcmp(menu[mop][copt], "Load")) load_picture();
	else if(!strcmp(menu[mop][copt], "New")){
		showmouse(0);
		clear_screen();
		showmouse(1);
		picname[0] = NULL;
	}else if(!strcmp(menu[mop][copt], "Exit")) done = 1;
	else if(!strcmp(menu[mop][copt], "Show Grid")){
		showmouse(0);
		grid_showing = 1 - grid_showing;
		draw_grid();
		update_grid();
		showmouse(1);
	}
	else if(!strcmp(menu[mop][copt], "Pencil")) drawmode = PENCIL;
	else if(!strcmp(menu[mop][copt], "Spill")) drawmode = SPILL;
	else if(!strcmp(menu[mop][copt], "Line")) drawmode = LINE;
	else if(!strcmp(menu[mop][copt], "Full Box")) drawmode = BOX_FULL;
	else if(!strcmp(menu[mop][copt], "Empty Box")) drawmode = BOX_EMPTY;
	else if(!strcmp(menu[mop][copt], "Airbrush")) drawmode = AIRBRUSH;
	else if(!strcmp(menu[mop][copt], "Paintbrush")) drawmode = PAINTBRUSH;
	else if(!strcmp(menu[mop][copt], "Replace")) drawmode = REPLACE;
	else if(!strcmp(menu[mop][copt], "Highlight")){
		drawmode = HIGHLIGHT;
		junk[0][0] = 24; junk[0][1] = NULL;
		junk[1][0] = 25; junk[1][1] = NULL;
		junk[2][0] = 27; junk[2][1] = NULL;
		junk[3][0] = 26; junk[3][1] = NULL;

		showmouse(0);
		highlight_direction = choose("Please choose a direction:", 4,
									  junk[0], 'H', junk[1], 'P',
									  junk[2], 'K', junk[3], 'M');
		showmouse(1);

	}else if(!strcmp(menu[mop][copt], "Envelop")) drawmode = ENVELOP;


	else if(!strcmp(menu[mop][copt], "Optimize Grid")) optimize_grid();
	else if(!strcmp(menu[mop][copt], "Grid Colour")){
		adjust_value("Colour:", &gridcolour, 0, 255);
		draw_grid();
	}else if(!strcmp(menu[mop][copt], "Air Speed")){
		adjust_value("Speed:", &airbrush_rate, 1, 100);
		adjust_value("Spread:", &airbrush_range, 1, 100);
	}else if(!strcmp(menu[mop][copt], "Load Palette")){
		showmouse(0);
		strcpy(palname, taketext("What would you like to call it?", &c, 2, "Okay", 13, "Cancel", 27));
		showmouse(1);
		if(!c){
			if(palette.load(palname))
				palette.apply();
			else{
				showmouse(0);
				choose("Unable to open file.", 1, "Okay", 13);
				showmouse(1);
			}
		}

	}

}

void load_picture(void){
	int a, b, arraysize, xsize, ysize;
	char *picture, c;
	FILE *fin = NULL;

	showmouse(0);
	strcpy(picname, taketext("What picture do you want to load?", &c, 2, "Okay", 13, "Cancel", 27));

	fin = fopen(picname, "rb");
	if(fin && !c){

		xsize = fgetc(fin);
		ysize = fgetc(fin);
		arraysize = xsize * ysize + 2;
		picture = (char *)malloc(arraysize);

		if(picture){

			picture[0] = xsize;
			picture[1] = ysize;
			fread((void *)&picture[2], 1, arraysize - 2, fin);

			if(picture[0] > xgridsize || picture[1] > ygridsize){
				if(!choose("Resize grid to fit picture?", 2, "Yes", 13, "No", 27)){
					a = picture[0];
					b = picture[1];
					if(a > 0 && a <= 100 && b > 0 && b <= 100){
						gridcolour = 0;
						draw_grid();
						gridcolour = DEFAULT_GRID_COLOUR;
						xgridsize = a;
						ygridsize = b;
						a = (palettex - gridx - 2) / xgridsize;
						b = (199 - gridy) / ygridsize;

						if(a > b) scale = b;
						else scale = a;

						box(gridx + 1, gridy + 1, gridx + xgridsize * scale - 2, gridy + ygridsize * scale - 2, 0, 0);
						draw_grid();
						update_grid();
					}else{
						choose("Picture size exceeds range (1-100)", 1, "Okay", 13);
					}
				}
			}

			cls();
			drawmenus(nummenus);
			draw_grid();
			draw_palette();
			put(picturex, picturey, picture, 0);
			update_grid();

			free(picture);
		}
	}

	if(fin) fclose(fin);

	showmouse(1);
}

void save_picture(char *filename){
	int x, y, arraysize;
	int x1 = 1000, y1 = 1000, x2 = 0, y2 = 0;
	char *picture, c;

	if(!strlen(filename)){
		showmouse(0);
		strcpy(filename, taketext("What would you like to call it?", &c, 2, "Okay", 13, "Cancel", 27));
		showmouse(1);
	}

	if(strlen(filename) > 0){
		strcpy(picname, filename);
		showmouse(0);
		for(x = picturex; x < picturex + xgridsize; x++){
			for(y = picturey; y < picturey + ygridsize; y++){
				if(vmem[x + 320 * y]){
					if(x < x1) x1 = x;
					if(y < y1) y1 = y;
					if(x > x2) x2 = x;
					if(y > y2) y2 = y;
				}
			}
		}
		if((x1 <= x2) && (y1 <= y2)){
			arraysize = (x2 - x1 + 1) * (y2 - y1 + 1) + 2;
			if((picture = (char *)malloc(arraysize)) != NULL){
				get(x1, y1, x2, y2, picture);
				savepic(picture, arraysize, filename);
			}else{
				sound(100);
				getch();
				nosound();
			}
		}
		showmouse(1);
	}
	if(picture) free(picture);
}

void clear_screen(void){
	cls();
	drawmenus(nummenus);
	draw_grid();
	draw_palette();
}

int on_palette(int x, int y){
	if((x >= palettex) && (x < palettex + 16 * palettescale) &&
	   (y >= palettey) && (y < palettey + 16 * palettescale))
		return 1;
	else return 0;
}

void draw_palette(void){
	int n;
	for(n = 0; n < 256; n++)
		box(palettex + (n % 16) * palettescale,
			palettey + ((int)(n / 16)) * palettescale,
			palettex + (n % 16 + 1) * palettescale,
			palettey + ((int)(n / 16 + 1)) * palettescale, n, 0);

	for(n = 0; n < 16; n++){
		box(palettex + palettescale * n, palettey,
			 palettex + palettescale * n, palettey + palettescale * 16, 8, 0);
		box(palettex, palettey + palettescale * n,
			 palettex + palettescale * 16, palettey + palettescale * n, 8, 0);
	}
	box(palettex, palettey, palettex + 16 * palettescale, palettey + 16 * palettescale, 8, 4);

	box(palettex + (colour % 16) * palettescale,
		palettey + (int)(colour / 16) * palettescale,
		palettex + (colour % 16 + 1) * palettescale,
		palettey + (int)(colour / 16 + 1) * palettescale, 15, 4);
}

int picx(int x){
	int returnval = -1;
	if((x >= gridx) && (x < gridx + xgridsize * scale))
		returnval = (int)((x - gridx - !grid_showing) / scale);
	return returnval;
}

int picy(int y){
	int returnval = -1;
	if((y >= gridy) && (y < gridy + ygridsize * scale))
		returnval = (int)((y - gridy - !grid_showing) / scale);
	return returnval;
}

void update_grid_area(int x1, int y1, int x2, int y2){
	int x, y;
	if(x1 > x2){ x1 ^= x2; x2 ^= x1; x1 ^= x2; }
	if(y1 > y2){ y1 ^= y2; y2 ^= y1; y1 ^= y2; }

	for(x = x1; x <= x2; x++){
		for(y = y1; y <= y2; y++){
			box(gridx + x * scale + 1, gridy + y * scale + 1,
				gridx + (x + 1) * scale - grid_showing,
				gridy + (y + 1) * scale - grid_showing,
				getpixel(picturex + x, picturey + y), 0);
		}
	}
}

void resize_grid(void){
	int a, b, x, y;
	char c;

	showmouse(0);
	a = atoi(taketext("What size X-axis?", &c, 1, "Okay"));
	b = atoi(taketext("What size Y-axis?", &c, 1, "Okay"));
	if(a > 0 && a <= 100 && b > 0 && b <= 100){
		gridcolour = 0;
		draw_grid();
		gridcolour = DEFAULT_GRID_COLOUR;
		xgridsize = a;
		ygridsize = b;
		a = (int)(palettex - gridx - 2) / xgridsize;
		b = (int)(199 - gridy) / ygridsize;
		if(b < a) scale = b;
		else scale = a;

		box(gridx + 1, gridy + 1, gridx + xgridsize * scale - 2, gridy + ygridsize * scale - 2, 0, 0);
		draw_grid();
		update_grid();
	}else{
		choose("Grid size exceeds range (1-100)", 1, "Okay", 13);
	}
	showmouse(1);
	update_pattern();
}

void optimize_grid(void){
	int x, y, x1, y1, x2, y2;

	showmouse(0);

	x2 = picturex;
	y2 = picturey;
	x1 = picturex + xgridsize;
	y1 = picturey + ygridsize;
	for(x = picturex; x < picturex + xgridsize; x++){
		for(y = picturey; y < picturey + ygridsize; y++){
			if(vmem[x + 320 * y]){
				if(x < x1) x1 = x;
				if(y < y1) y1 = y;
				if(x > x2) x2 = x;
				if(y > y2) y2 = y;
			}
		}
	}

	gridcolour = 0;
	draw_grid();
	gridcolour = DEFAULT_GRID_COLOUR;

	xgridsize = x2 - x1 + 1;
	ygridsize = y2 - y1 + 1;

	x = (int)(palettex - gridx - 2) / xgridsize;
	y = (int)(199 - gridy) / ygridsize;
	if(y < x) scale = y;
	else scale = x;
	draw_grid();
	update_grid();
	showmouse(1);
}

void update_grid(void){
	int x, y;

	for(x = 0; x < xgridsize; x++){
		for(y = 0; y < ygridsize; y++){
			box(gridx + x * scale + 1, gridy + y * scale + 1,
				gridx + (x + 1) * scale - grid_showing,
				gridy + (y + 1) * scale - grid_showing,
				getpixel(picturex + x, picturey + y), 0);
		}
	}
}

void update_pattern(void){
	int maxgx, maxgy, maxpx, maxpy, xoff, yoff, minpx, mingy, mingx, minpy;
	unsigned char *pic;
	pic = (unsigned char *)malloc(2 + xgridsize * ygridsize);

	if(pic){
		get(picturex, picturey, picturex + xgridsize - 1, picturey + ygridsize - 1, pic);

		maxgx = gridx + 1 + xgridsize * scale + 1 - grid_showing;
		maxgy = gridy + 1 + ygridsize * scale + 1 - grid_showing;
		mingy = gridy - 1;
		mingx = gridx - 1;
		minpx = palettex - 1;
		minpy = palettey - 1;
		maxpx = palettex + 16 * palettescale + 1;
		maxpy = palettey + 16 * palettescale + 1;
		xoff = xgridsize - picturex % xgridsize;
		yoff = ygridsize - picturey % ygridsize;

		erase(maxmenux + 1, 0, minpx - 1, maxmenuy - 1, xoff, yoff, pic);
		erase(minpx + 1, 0, maxpx - 1, minpy, xoff, yoff, pic);
		erase(0, mingy + 1, mingx, maxgy - 1, xoff, yoff, pic);
		erase(0, maxmenuy, maxgx, mingy, xoff, yoff, pic);
		erase(maxgx, maxpy, 319, maxgy, xoff, yoff, pic);
		erase(0, maxgy, 319, 199, xoff, yoff, pic);
		erase(maxpx, 0, 319, maxpy, xoff, yoff, pic);
		erase(maxgx, 0, minpx, maxpy, xoff, yoff, pic);
		free(pic);
	}
}


void draw_grid(void){
	int n;
	if(grid_showing){
		box(gridx, gridy,
			gridx + 1 + xgridsize * scale,
			gridy + 1 + ygridsize * scale, 0, 4);

		for(n = 0; n < xgridsize; n++)
			box(gridx + n * scale, gridy, gridx + n * scale, gridy + ygridsize * scale, gridcolour, 0);
		for(n = 0; n < ygridsize; n++)
			box(gridx, gridy + n * scale, gridx + xgridsize * scale, gridy + n * scale, gridcolour, 0);
		box(gridx, gridy, gridx + xgridsize * scale, gridy + ygridsize * scale, gridcolour, 4);
	}else{
		box(gridx, gridy,
			gridx + 1 + xgridsize * scale,
			gridy + 1 + ygridsize * scale, gridcolour, 4);
	}
}

void handle_spill(int px, int py){
	unsigned char c = colour;
	if(buttonpressed() == 2) c = 0;

	showmouse(0);
	if(getpixel(picturex + px, picturey + py) != c){
		spill(picturex + px, picturey + py, c);
		update_grid();
		update_pattern();
	}
	showmouse(1);
}

void handle_pencil(int px, int py){
	unsigned char c = colour;
	if(buttonpressed() == 2) c = 0;

	showmouse(0);
	box(gridx + px * scale + 1, gridy + py * scale + 1,
		gridx + (px + 1) * scale - grid_showing,
		gridy + (py + 1) * scale - grid_showing,
		c, 0);
	putpixel(picturex + px, picturey + py, c, 0);
	update_pattern();
	showmouse(1);
}

void handle_line(int px, int py){
	int x1, y1, x2, y2, ux1, uy1, ux2, uy2;
	unsigned char c = colour;

	if(buttonpressed() == 2) c = 0;

	setmouselimits(gridx, gridy,
				   gridx + xgridsize * scale - 1,
				   gridy + ygridsize * scale -1);
	x2 = x1 = px;
	y2 = y1 = py;
	showmouse(0);
	line(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
	showmouse(1);
	while(buttonpressed()){
		px = picx(mousex());
		py = picy(mousey());
		if(px != x2 || py != y2){
			ux1 = x1;
			uy1 = y1;
			ux2 = x2;
			uy2 = y2;
			if(ux1 > ux2){ ux1 ^= ux2; ux2 ^= ux1; ux1 ^= ux2; }
			if(uy1 > uy2){ uy1 ^= uy2; uy2 ^= uy1; uy1 ^= uy2; }
			if(px < ux1) ux1 = px;
			if(px > ux2) ux2 = px;
			if(py < uy1) uy1 = py;
			if(py > uy2) uy2 = py;

			showmouse(0);
			line(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
			x2 = px;
			y2 = py;
			line(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
			update_pattern();
			update_grid_area(ux1, uy1, ux2, uy2);
			showmouse(1);
		}
	}
	showmouse(0);
	line(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c, 0);
	update_grid();
	update_pattern();
	showmouse(1);
	setmouselimits(0, 0, 319, 199);
}

void handle_box(int px, int py){
	int x1, y1, x2, y2, ux1, uy1, ux2, uy2;
	unsigned char c = colour;

	setmouselimits(gridx, gridy,
				   gridx + xgridsize * scale - 1,
				   gridy + ygridsize * scale -1);
	x2 = x1 = px;
	y2 = y1 = py;
	showmouse(0);
	line(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
	showmouse(1);
	while(buttonpressed()){
		px = picx(mousex());
		py = picy(mousey());
		if(px != x2 || py != y2){
			ux1 = x1;
			uy1 = y1;
			ux2 = x2;
			uy2 = y2;
			if(ux1 > ux2){ ux1 ^= ux2; ux2 ^= ux1; ux1 ^= ux2; }
			if(uy1 > uy2){ uy1 ^= uy2; uy2 ^= uy1; uy1 ^= uy2; }
			if(px < ux1) ux1 = px;
			if(px > ux2) ux2 = px;
			if(py < uy1) uy1 = py;
			if(py > uy2) uy2 = py;

			showmouse(0);
			box(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
			x2 = px;
			y2 = py;
			box(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
			update_pattern();
			update_grid_area(ux1, uy1, ux2, uy2);
			showmouse(1);
		}
	}
	showmouse(0);
	box(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c, 0);
	update_grid();
	showmouse(1);
	setmouselimits(0, 0, 319, 199);
}

void handle_empty_box(int px, int py){
	int x1, y1, x2, y2, ux1, uy1, ux2, uy2;
	unsigned char c = colour;

	setmouselimits(gridx, gridy,
				   gridx + xgridsize * scale - 1,
				   gridy + ygridsize * scale -1);
	x2 = x1 = px;
	y2 = y1 = py;
	showmouse(0);
	line(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
	showmouse(1);
	while(buttonpressed()){
		px = picx(mousex());
		py = picy(mousey());
		if(px != x2 || py != y2){
			ux1 = x1;
			uy1 = y1;
			ux2 = x2;
			uy2 = y2;
			if(ux1 > ux2){ ux1 ^= ux2; ux2 ^= ux1; ux1 ^= ux2; }
			if(uy1 > uy2){ uy1 ^= uy2; uy2 ^= uy1; uy1 ^= uy2; }
			if(px < ux1) ux1 = px;
			if(px > ux2) ux2 = px;
			if(py < uy1) uy1 = py;
			if(py > uy2) uy2 = py;

			showmouse(0);

			box(picturex + x1, picturey + y1, picturex + x1, picturey + y2, c + !c, 1);
			box(picturex + x2, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
			box(picturex + x1, picturey + y1, picturex + x2, picturey + y1, c + !c, 1);
			box(picturex + x1, picturey + y2, picturex + x2, picturey + y2, c + !c, 1);
			x2 = px;
			y2 = py;
			box(picturex + x1, picturey + y1, picturex + x1, picturey + y2, c + !c, 1);
			box(picturex + x2, picturey + y1, picturex + x2, picturey + y2, c + !c, 1);
			box(picturex + x1, picturey + y1, picturex + x2, picturey + y1, c + !c, 1);
			box(picturex + x1, picturey + y2, picturex + x2, picturey + y2, c + !c, 1);
			update_pattern();
			update_grid_area(ux1, uy1, ux2, uy2);
			showmouse(1);
		}
	}
	showmouse(0);
	box(picturex + x1, picturey + y1, picturex + x2, picturey + y2, c, 4);
	update_grid();
	showmouse(1);
	setmouselimits(0, 0, 319, 199);
}

void handle_airbrush(void){
	unsigned int count;
	int x, y, n, gx, gy;
	unsigned char c = colour;

	if(buttonpressed() == 2) c = 0;


	showmouse(0);
	while(buttonpressed()){
		while(count == counter());
		count = counter();
		for(n = 0; n < airbrush_rate; n++){
			x = rand() % airbrush_range;
			y = rand() % airbrush_range;
			x -= airbrush_range >> 1;
			y -= airbrush_range >> 1;

			x += (mousex() - gridx - 1) / scale;
			y += (mousey() - gridy - 1) / scale;
			while(x < 0) x += xgridsize;
			while(y < 0) y += ygridsize;
			x %= xgridsize;
			y %= ygridsize;
			putpixel(picturex + x, picturey + y, c, 0);
			box(gridx + x * scale + 1, gridy + y * scale + 1,
				gridx + (x + 1) * scale - grid_showing,
				gridy + (y + 1) * scale - grid_showing,
				c, 0);
		}
		update_pattern();
	}
	showmouse(1);
}

void handle_paintbrush(void){
	int mx, my, x1, y1, x2, y2, oldmx, oldmy;
	unsigned char c = colour;

	readmousepos(&mx, &my);
	oldmx = mx;
	oldmy = my;

	if(buttonpressed() == 2) c = 0;

	setmouselimits(gridx, gridy,
				   gridx + xgridsize * scale - 1,
				   gridy + ygridsize * scale - 1);
	while(buttonpressed()){
		readmousepos(&mx, &my);
		x1 = picx(oldmx);
		y1 = picy(oldmy);
		x2 = picx(mx);
		y2 = picy(my);

		if(x1 != x2 || y1 != y2){
			oldmx = mx;
			oldmy = my;
			showmouse(0);
			line(picturex + x1, picturey + y1,
				 picturex + x2, picturey + y2, c, 0);
			update_grid_area(x1, y1, x2, y2);
			update_pattern();
			showmouse(1);
		}
	}
	setmouselimits(0, 0, 319, 199);
}

void handle_replacement(void){
	int x, y, dx, dy;
	unsigned char c, sample;
	c = colour;
	if(buttonpressed() == 2) c = 0;
	showmouse(0);
	sample = getpixel(picturex + picx(mousex()), picturey + picy(mousey()));
	for(y = 0; y < ygridsize; y++){
		for(x = 0; x < xgridsize; x++){
			dx = x + picturex;
			dy = y + picturey;
			if(getpixel(dx, dy) == sample) putpixel(dx, dy, c, 0);
		}
	}
	update_grid();
	update_pattern();
	showmouse(1);
}

void handle_envelopment(void){
	int x, y, dx, dy, px, py;
	int drawx, drawy;
	unsigned char c, sample;
	c = colour;
	if(buttonpressed() == 2) c = 0;
	showmouse(0);
	sample = getpixel(picturex + picx(mousex()), picturey + picy(mousey()));
	for(y = 0; y < ygridsize; y++){
		for(x = 0; x < xgridsize; x++){
			px = x + picturex;
			py = y + picturey;
			if(getpixel(px, py) == sample){
				for(dx = -1; dx < 2; dx++){
					for(dy = -1; dy < 2; dy++){
						if(!dx ^ !dy){
							drawx = x + dx;
							drawy = y + dy;
							while(drawx < 0) drawx += xgridsize;
							while(drawy < 0) drawy += ygridsize;
							drawx %= xgridsize;
							drawy %= ygridsize;
							drawx += picturex;
							drawy += picturey;

							if(!getpixel(drawx, drawy) &&
							   getpixel(drawx, drawy) != sample)

								putpixel(drawx, drawy, c, 0);

						}
					}
				}

			}
		}
	}
	update_grid();
	update_pattern();
	showmouse(1);
}

void handle_highlighting(int mx, int my){
	int x, y, xoffset, yoffset, drawx, drawy;
	unsigned char c, sample;
	c = colour;
	if(buttonpressed() == 2) c = 0;

	switch(highlight_direction){
		case 0: /* up */	xoffset = 0; 	yoffset = -1; 	break;
		case 1: /* down */	xoffset = 0; 	yoffset = 1; 	break;
		case 2: /* left */	xoffset = -1; 	yoffset = 0; 	break;
		case 3: /* right */	xoffset = 1; 	yoffset = 0;
	}

	showmouse(0);
	sample = getpixel(picturex + mx, picturey + my);
	if(c != sample){
		for(x = 0; x < xgridsize; x++){
			for(y = 0; y < ygridsize; y++){

				if(getpixel(picturex + x, picturey + y) == sample){
					drawx = x + xoffset;
					drawy = y + yoffset;
					while(drawx < 0) drawx += xgridsize;
					while(drawy < 0) drawy += ygridsize;
					drawx %= xgridsize;
					drawy %= ygridsize;
					drawx += picturex;
					drawy += picturey;
					if(getpixel(drawx, drawy) != sample)
						putpixel(drawx, drawy, c, 0);
				}
			}
		}
	}

	update_grid();
	update_pattern();
	showmouse(1);
}


void handle_drawing(int px, int py){
	int x1, y1, x2, y2, ux1, ux2, uy1, uy2;

	switch(drawmode){
		case PENCIL: handle_pencil(px, py); break;
		case SPILL: handle_spill(px, py); break;
		case LINE: handle_line(px, py); break;
		case BOX_FULL: handle_box(px, py); break;
		case BOX_EMPTY: handle_empty_box(px, py); break;
		case AIRBRUSH: handle_airbrush(); break;
		case PAINTBRUSH: handle_paintbrush(); break;
		case REPLACE: handle_replacement(); break;
		case ENVELOP: handle_envelopment(); break;
		case HIGHLIGHT: handle_highlighting(px, py); break;
	}
	showmouse(1);
}

int current_option(int my, int menuy, int menulength){
	int n = 0, y;
	for(y = menuy + 2; (n <= menulength) && (y < my); n++)
		y += yfontsize;

	if((n > menulength) || (n == 0)) return 0;
	else return n;
}

int on_open_menu(int mx, int my, int menux, int menuy,
				 int menuxsize, int menuysize)
{
	int returnval = 0;
	if((mx >= menux) && (mx < menux + menuxsize))
		if((my >= menuy) && (my < menuy + menuysize))
			returnval = 1;
	return returnval;
}

void close_menu(char *undermenu, int menux, int menuy, int *menu_open){
	*menu_open = -1;
	showmouse(0);
	put(menux, menuy, undermenu, 0);
	showmouse(1);
}

void open_menu(int cm, int *menulength, char *undermenu, int *mnux, int *mnuy){
	int n, xsize = 0, ysize, menux = (int)(xfontsize / 2), menuy = 0, x, y;

	/*** determine X size of menu ***/
	for(n = 1; n <= menulength[cm]; n++)
		if(strlen(menu[cm][n]) > xsize) xsize = strlen(menu[cm][n]);
	xsize *= xfontsize;
	xsize += 6;

	/*** determine Y size of menu ***/
	ysize = (menulength[cm] * yfontsize) + 6;

	/*** determine X, Y locations of the menu ***/
	for(n = 0; n < cm; n++)
		menux += xfontsize * (strlen(menu[n][0]));
	menuy = yfontsize;

	/*** draw the menu ***/
	get(menux, menuy, menux + xsize - 1, menuy + ysize - 1, undermenu);
	box(menux, menuy, menux + xsize - 1, menuy + ysize - 1, 0, 0);
	x = menux + 3;
	y = menuy + 3;
	for(n = 1; n <= menulength[cm]; n++){
		drawtext(x, y, menu[cm][n], textcolour, 0);
		y += yfontsize;
	}
	box(menux, menuy, menux, menuy + ysize - 1, 28, 0);
	box(menux, menuy, menux + xsize - 1, menuy, 28, 0);
	box(menux + xsize - 1, menuy, menux + xsize - 1, menuy + ysize - 1, 18, 0);
	box(menux, menuy + ysize - 1, menux + xsize - 1, menuy + ysize - 1, 18, 0);
	box(menux, menuy, menux + xsize - 1, menuy + ysize - 1, 22, 5);
	*mnux = menux;
	*mnuy = menuy;
}

int current_menu(int mx, int my, int nummenus){
	int n, xmenusize = 3, ymenusize = yfontsize + 6, returnval = 0;

	if(my <= ymenusize){
		for(n = 0; n < nummenus; n++){
			xmenusize += xfontsize * (strlen(menu[n][0]));
			if((xmenusize - xfontsize / 2 + 1) < mx) returnval++;
		}
		if(returnval == nummenus) returnval = -1;
	}
	else returnval = -1;

	return returnval;
}

void drawmenus(int nummenus){
	int n, xmenusize = menubordersize;
	int ymenusize = yfontsize + 2 * menubordersize;

	for(n = 0; n < nummenus; n++){
		box(xmenusize - xfontsize / 2 + 1, 0, xmenusize - xfontsize / 2 + 1, ymenusize - 1, 26, 0);
		drawtext(xmenusize, menubordersize, menu[n][0], textcolour, 0);
		xmenusize += xfontsize * (strlen(menu[n][0]));
		box(xmenusize - xfontsize / 2, 0, xmenusize - xfontsize / 2 , ymenusize - 1, 18, 0);
	}
	xmenusize -= menubordersize;

	box(xmenusize - 1, 1, xmenusize - 1, ymenusize - 1, 18, 5);
	box(1, ymenusize - 1, xmenusize - 1, ymenusize - 1, 18, 5);
	box(0, 0, xmenusize - 1, 0, 26, 5);
	box(0, 0, 0, ymenusize - 1, 26, 5);
	box(0, 0, xmenusize - 1, ymenusize - 1, 22, 5);
}

void load_menus(unsigned char *filename, int *nummenus, int *menulength){
	int n, m;
	char junkstring[maxmenuwidth];
	FILE *menufile;

	/***** get the number of menus ********/
	menufile = fopen(filename, "r");
	*nummenus = atoi(fgets(junkstring, maxmenuwidth, menufile));

	/***** load each menu ********/
	for(n = 0; n < *nummenus; n++){
		fgets(menu[n][0], maxmenuwidth, menufile);
		menulength[n] = atoi(fgets(junkstring, maxmenuwidth, menufile));
		for(m = 1; m <= menulength[n]; m++){
			fgets(menu[n][m], maxmenuwidth, menufile);
			menu[n][m][strlen(menu[n][m]) - 1] = (char)0;
		}
	}
	fclose(menufile);
}


/**************************** Keyboard I/O *****************************/
void set_keyboard_rate(unsigned char rate, unsigned char delay){
	delay %= 4;
	rate %= 32;
	asm{
		mov ah, 0x3
		mov al, 0x5
		mov bh, delay
		mov bl, rate
		int 0x16
	}
}

unsigned char readkbd(void){
	if(!kbhit()) asm int 9; /* kills keyboard delay */
	if(kbhit())
		return getch();
	else
		return NULL;
}

char blackcharstatus(void){
	char returnval;
	asm{
		mov ah, 0x2
		int 0x16
		mov returnval, al
	}
	return returnval;
}

char kbscan(void){
	return inp(0x60);
}

unsigned char getch(void){
	unsigned char returnval;
	asm{
		mov ah, 0x8
		int 0x21
		mov returnval, al
	}
	return returnval;
}

unsigned char kbhit(void){
	unsigned char returnval;
	asm{
		mov ah, 0xB
		int 0x21
		mov returnval, al
	}
	returnval = returnval != 0;
	return returnval;
}

void drawchar(unsigned char c){
	asm{
		mov ah, 0x2
		mov dl, c
		int 0x21
	}
}

/**************************** Mouse I/O *****************************/
void readmousediff(void){
	int x, y;
	asm{
		mov ax, 0xB
		int 0x33
		mov x, cx
		mov y, dx
	}
	mousediff.deltax = x;
	mousediff.deltay = y;
}

char buttonpressed(void){
	int returnval;
	asm{
		mov ax, 0x3
		int 0x33
		mov returnval, bx
	}
	return returnval;
}

void readmousepos(int *mx, int *my){
	int x, y;

	asm{
		mov ax, 0x3
		int 0x33
		mov x, cx
		mov y, dx
	}

	*mx = x >> 1;
	*my = y;
}

int mousex(void){
	int returnval;
	asm{
		mov ax, 0x3
		int 0x33
		mov returnval, cx
	}
	return returnval >> 1;
}

int mousey(void){
	int returnval;
	asm{
		mov ax, 0x3
		int 0x33
		mov returnval, dx
	}
	return returnval;

}

void initmouse(void){
	asm{
		mov ax, 0
		int 0x33
	}
}

void showmouse(int onoff){
	onoff = (onoff == 0) + 1;
	asm{
		mov ax, onoff
		int 0x33
	}
}

char turnoffmouse(void){
	int returnval;
	asm{
		mov ax, 0x1F
		int 0x33
		mov returnval, ax
	}
	return returnval;
}

void setmousepos(int x, int y){
	asm{
		mov ax, 0x4
		mov cx, x
		shl cx, 1
		mov dx, y
		int 0x33
	}
}

void setmouselimits(int x1, int y1, int x2, int y2){
	asm{
		/*********** set x range **********/
		mov ax, 0x7
		mov cx, x1
		mov dx, x2
		shl cx, 1
		shl dx, 1
		int 0x33

		/*********** set y range **********/
		mov ax, 0x8
		mov cx, y1
		mov dx, y2
		int 0x33
	}
}

/****************************** Misc ******************************/
void locate(char x, char y){
	asm{
		mov ah, 2
		mov dh, y
		mov dl, x
		mov bh, 0
		int 0x10
	}
}

char sgn(int val){
	if(val > 0) return 1;
	return -(val < 0);
}

unsigned int counter(void){
	return peek(0x40, 0x6C);
}

void set_count_rate(int rate){
	_asm cli
	if(rate < 19) rate = 19;
	rate = (rate == 18) ? 0 : ((unsigned int) (1193180L / ((unsigned long)rate)));
	asm{
		mov dx, 0x43
		mov al, 0x34
		out dx, al
		mov dx, 0x40
		mov ax, rate
		out dx, al
		mov al, ah
		out dx, al
		sti
	}
}

/****************************** Graphics ******************************/
void initscrn(char mode){
	asm{
		mov ah, 0
		mov al, mode
		int 0x10
	}
}

unsigned char csmode(void){
	unsigned char returnval;
	asm{
		mov ah, 0xF
		int 0x10
		mov returnval, al
	}
	return returnval;
}

void putpixel(int x, int y, unsigned char colour, char mode){
	if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
		switch(mode){
			case 0: vmem[(y << 8) + (y << 6) + x]  = colour; break;
			case 1: vmem[(y << 8) + (y << 6) + x] ^= colour; break;
			case 2: vmem[(y << 8) + (y << 6) + x] |= colour; break;
			case 3: vmem[(y << 8) + (y << 6) + x] &= colour; break;
			case 4:
				if(colour) vmem[(y << 8) + (y << 6) + x] = colour;
				break;
			case 5:
				if(!vmem[(y << 8) + (y << 6) + x])
					vmem[(y << 8) + (y << 6) + x] = colour;
				break;
			case 6: vmem[(y << 8) + (y << 6) + x] += colour; break;
		}
}

char getpixel(int x, int y){
	if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
		return vmem[(y << 8) + (y << 6) + x];
	else
		return 0;
}

void line(int x1, int y1, int x2, int y2, unsigned char colour, char mode){
	int n, deltax, deltay, sgndeltax, sgndeltay, deltaxabs, deltayabs, x, y, drawx, drawy;

	deltax = x2 - x1;
	deltay = y2 - y1;
	deltaxabs = abs(deltax);
	deltayabs = abs(deltay);
	sgndeltax = sgn(deltax);
	sgndeltay = sgn(deltay);
	x = deltayabs >> 1;
	y = deltaxabs >> 1;
	drawx = x1;
	drawy = y1;

	putpixel(drawx, drawy, colour, mode);

	if(deltaxabs >= deltayabs){
		for(n = 0; n < deltaxabs; n++){
			y += deltayabs;
			if(y >= deltaxabs){
				y -= deltaxabs;
				drawy += sgndeltay;
			}
			drawx += sgndeltax;
			putpixel(drawx, drawy, colour, mode);
		}
	}else{
		for(n = 0; n < deltayabs; n++){
			x += deltaxabs;
			if(x >= deltayabs){
				x -= deltayabs;
				drawx += sgndeltax;
			}
			drawy += sgndeltay;
			putpixel(drawx, drawy, colour, mode);
		}
	}
}

void box(int x1, int y1, int x2, int y2, int colour, char mode){
	int x, y, n;
	if(x1 < 0) x1 = 0; if(x1 > 319) x1 = 319;
	if(y1 < 0) y1 = 0; if(y1 > 199) y1 = 199;
	if(x2 < 0) x2 = 0; if(x2 > 319) x2 = 319;
	if(y2 < 0) y2 = 0; if(y2 > 199) y2 = 199;
	if(x1 > x2) { x1 ^= x2; x2 ^= x1; x1 ^= x2; }
	if(y1 > y2) { y1 ^= y2; y2 ^= y1; y1 ^= y2; }

	if(mode == 4){
		for(x = x1; x <= x2; x++){
			vmem[(y1 << 8) + (y1 << 6) + x] = colour;
			vmem[(y2 << 8) + (y2 << 6) + x] = colour;
		}
		for(y = y1; y <= y2; y++){
			vmem[(y << 8) + (y << 6) + x1] = colour;
			vmem[(y << 8) + (y << 6) + x2] = colour;
		}

	}else{
		for(x = x1; x <= x2; x++){
			for(y = y1; y <= y2; y++){
				switch(mode){
					case 0: vmem[(y << 8) + (y << 6) + x]  = colour; break;
					case 1:	vmem[(y << 8) + (y << 6) + x] ^= colour; break;
					case 2: vmem[(y << 8) + (y << 6) + x] |= colour; break;
					case 3: vmem[(y << 8) + (y << 6) + x] &= colour; break;
					case 5:
						n = (y << 8) + (y << 6) + x;
						vmem[n] += (!vmem[n]) * colour;
					break;
					case 6: vmem[(y << 8) + (y << 6) + x] += colour; break;
				}
			}
		}
	}
}

void circle(int cx, int cy, int radius, unsigned char colour, char mode){
	int ydist, ydist2, xdist, x, y, dx2;
	int x1;
	if(radius < 1) radius *= -1;
	if(!radius) radius++;
	x1 = cx - radius;
	for(x = x1; x <= cx; x += 1){
		xdist = abs(x - cx);
		ydist = sqrt(radius * radius - xdist * xdist);
		ydist2 = sqrt(radius * radius - (xdist - 1) * (xdist - 1));
		ydist2 += !xdist;
		if(mode == 4){
			if(abs(ydist2 - ydist) > 1){
				for(y = ydist; y < ydist2; y++){
					if((x >= 0) && (x <= 319)){
						if((cy + y >= 0) && (cy + y <= 199)) vmem[x + 320 * (cy + y)] = colour;
						if((cy - y >= 0) && (cy - y <= 199)) vmem[x + 320 * (cy - y)] = colour;
					}
					dx2 = cx + xdist;
					if((dx2 >= 0) && (dx2 <= 319)){
						if((cy + y >= 0) && (cy + y <= 199)) vmem[dx2 + 320 * (cy + y)] = colour;
						if((cy - y >= 0) && (cy - y <= 199)) vmem[dx2 + 320 * (cy - y)] = colour;
					}
				}
			}else{
				dx2 = cx + xdist;
				y = cy + ydist;
				if((y >= 0) && (y <= 199)){
					if((x >= 0) && (x <= 319)) vmem[x + 320 * y] = colour;
					if((dx2 >= 0) && (dx2 <= 319)) vmem[dx2 + 320 * y] = colour;
				}
				y = cy - ydist;
				if((y >= 0) && (y <= 199)){
					if((x >= 0) && (x <= 319)) vmem[x + 320 * y] = colour;
					if((dx2 >= 0) && (dx2 <= 319)) vmem[dx2 + 320 * y] = colour;
				}
			}
		}else{
			box(x, cy + ydist2, x, cy - ydist2, colour, mode);
			dx2 = cx + xdist;
			if(xdist > 0)box(dx2, cy + ydist2, dx2, cy - ydist2, colour, mode);
		}
	}
}

void spill(int spillx, int spilly, int spillcolour){
	int x, y = spilly, doneline = 0, backgcolour;
	int minspillx, maxspillx;

	backgcolour = getpixel(spillx, spilly);
	if(backgcolour != spillcolour){
		for(x = spillx; !doneline; x--){
			if(x <= picturex) doneline = 1;
			if(getpixel(x, y) == backgcolour) minspillx = x;
			else doneline = 1;
		}

		doneline = 0;
		for(x = spillx; !doneline; x++){
			if(x >= picturex + xgridsize - 1) doneline = 1;
			if(getpixel(x, y) == backgcolour) maxspillx = x;
			else doneline = 1;
		}

		line(minspillx, y, maxspillx, y, spillcolour, 0);
		for(x = minspillx; x <= maxspillx; x++){
			if(y - 1 >= picturey)
				if(getpixel(x, y - 1) == backgcolour) spill(x, y - 1, spillcolour);
			if(y + 1 < picturey + ygridsize)
				if(getpixel(x, y + 1) == backgcolour) spill(x, y + 1, spillcolour);
		}
	}
}

void drawpoly(int *x, int *y, unsigned char colour, char mode){
	int dx, mny, mxy, mnx, mxx, yc;
	int mul1, div1, mul2, div2, mul3, div3, mul4, div4;

	// find the maximum y (mxy) and minimum y (mny)
	mny = y[0];
	mxy = y[0];
	if(y[1] < mny) mny = y[1];
	if(y[1] > mxy) mxy = y[1];
	if(y[2] < mny) mny = y[2];
	if(y[2] > mxy) mxy = y[2];
	if(y[3] < mny) mny = y[3];
	if(y[3] > mxy) mxy = y[3];

	// if the mimimum or maximum is out of bounds, bring it back in
	if(mny < 0) mny = 0;
	if(mxy > 199) mxy = 199;

	// verticle range checking
	if(mny > 199) return;
	if(mxy < 0) return;

	// constants needed for intersection calculations
	mul1 = x[0] - x[3];  div1 = y[0] - y[3];
	mul2 = x[1] - x[0];  div2 = y[1] - y[0];
	mul3 = x[2] - x[1];  div3 = y[2] - y[1];
	mul4 = x[3] - x[2];  div4 = y[3] - y[2];

	for(yc = mny; yc < mxy; yc++){
		mnx = 320;
		mxx = -1;

		if(((y[3] >= yc) || (y[0] >= yc))
		 && ((y[3] <= yc) || (y[0] <= yc))
		 && (y[3] != y[0]))
		{
			dx = ((yc - y[3]) * mul1 / div1) + x[3];
			if(dx < mnx) mnx = dx;
				if(dx > mxx) mxx = dx;
		}

		if(((y[0] >= yc) || (y[1] >= yc))
		 && ((y[0] <= yc) || (y[1] <= yc))
		 && (y[0] != y[1]))
		{
			dx = ((yc - y[0]) * mul2 / div2) + x[0];
			if(dx < mnx) mnx = dx;
			if(dx > mxx) mxx = dx;
		}

		if(((y[1] >= yc) || (y[2] >= yc))
		 && ((y[1] <= yc) || (y[2] <= yc))
		 && (y[1] != y[2]))
		{
			dx = ((yc - y[1]) * mul3 / div3) + x[1];
			if(dx < mnx) mnx = dx;
			if(dx > mxx) mxx = dx;
		}

		if(((y[2] >= yc) || (y[3] >= yc))
		 &&((y[2] <= yc) || (y[3] <= yc))
		 &&(y[2] != y[3]))
		{
			dx = ((yc - y[2]) * mul4 / div4) + x[2];
			if(dx < mnx) mnx = dx;
			if(dx > mxx) mxx = dx;
		}

		// horizontal range checking
		if(mnx < 0) mnx = 0;
		if(mxx > 319) mxx = 319;

		if(mnx <= mxx)
			hline(mnx, mxx, yc, colour, mode);
	}
}

void hline(int x1, int x2, int y, unsigned char colour, char mode){
	int n;
	if(!mode)
		_fmemset(&vmem[x1 + 320 * y], colour, x2 - x1 + 1);
	else
		for(n = x1; n <= x2; n++)
			putpixel(n, y, colour, mode);
}

void set_colour(int c, char r, char g, char b){
	r %= 64;
	g %= 64;
	b %= 64;
	asm{
		mov ah, 0x10
		mov al, 0x10
		mov bx, c
		mov ch, g
		mov cl, b
		mov dh, r
		int 0x10
	}
}

void cls(void){
	_fmemset(vmem, 0, 64000);
}

void drawtext(int x, int y, unsigned char *string, unsigned char colour, char mode){
	int length = strlen(string), n, m, initx = x;
	char charval, junk[100], c, repeat;

	for(n = 0; n < length; n++){
		do{
			charval = string[n];
			repeat = 1;
			switch(charval){
				case '\r':
					x = initx; /** please note: break is gone for a reason. */
				case '\n':
					y += yfontsize;
					break;
				case '\t':
					x += tabsize * xfontsize;
					break;
				case '\b':
					x -= xfontsize;
					break;
				case '\v':
					y += 4 * yfontsize;
					break;
				default:
					repeat = 0;
			}
			n += repeat;
		}while(repeat);
		putfont(x, y, font[charval], colour, mode);
		x += xfontsize;
	}
}

void get(int x1, int y1, int x2, int y2, unsigned char far *picture){
	int n = 2, x, y;
	if(x1 > x2){ x1 ^= x2; x2 ^= x1; x1 ^= x2; }
	if(y1 > y2){ y1 ^= y2; y2 ^= y1; y1 ^= y2; }

	picture[0] = x2 - x1 + 1;
	picture[1] = y2 - y1 + 1;
	for(y = y1; y <= y2; y++){
		for(x = x1; x <= x2; x++){
			if((x >= 0) || (x <= 319) || (y >= 0) || (y <= 199))
				picture[n] = vmem[(y << 8) + (y << 6) + x];
			else
				picture[n] = 0;

			n++;
		}
	}
}

void putfont(int x1, int y1, unsigned char far *picture, unsigned char colour, char mode){
	int n = 2, x, y;
	unsigned int dy, madd;
	int x2 = x1 + picture[0] - 1;
	int y2 = y1 + picture[1] - 1;
	mode %= 7;
	for(y = y1; y <= y2; y++){
		if(y >= 0 && y <= 199){
			dy = (y << 8) + (y << 6);
			for(x = x1; x <= x2; x++){
				if((x >= 0) && (x <= 319)){
					madd = dy + x;
					switch(mode){
						case 0: vmem[madd]  = (picture[n] != 0) * colour; break;
						case 1: vmem[madd] ^= (picture[n] != 0) * colour; break;
						case 2: vmem[madd] |= (picture[n] != 0) * colour; break;
						case 3: vmem[madd] &= (picture[n] != 0) * colour; break;
						case 4:
							if(picture[n] != 0)
								vmem[madd] = (picture[n] != 0) * colour;
							break;
						case 5:
							if(vmem[madd] == 0)
								vmem[madd] = (picture[n] != 0) * colour;
							break;
						case 6: vmem[madd] += (picture[n] != 0) * colour; break;
					}
				}
				n++;
			}
		}
	}
}

void put(int x1, int y1, unsigned char far *picture, char mode){
	int n = 2, x, y;
	int x2 = x1 + picture[0] - 1;
	int y2 = y1 + picture[1] - 1;

	switch(mode){
		case 0:
			for(y = y1; y <= y2; y++)
				for(x = x1; x <= x2; x++){
					if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
						vmem[(y << 8) + (y << 6) + x]  = picture[n];
					n++;
				}
		break;
		case 1:
			for(y = y1; y <= y2; y++)
				for(x = x1; x <= x2; x++){
					if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
						vmem[(y << 8) + (y << 6) + x] ^= picture[n];
					n++;
				}
		break;
		case 2:
			for(y = y1; y <= y2; y++)
				for(x = x1; x <= x2; x++){
					if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
						vmem[(y << 8) + (y << 6) + x] |= picture[n];
					n++;
				}
		break;
		case 3:
			for(y = y1; y <= y2; y++)
				for(x = x1; x <= x2; x++){
					if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
						vmem[(y << 8) + (y << 6) + x] &= picture[n];
					n++;
				}
		break;
		case 4:
			for(y = y1; y <= y2; y++)
				for(x = x1; x <= x2; x++){
					if(picture[n] != 0)
						if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
							vmem[(y << 8) + (y << 6) + x] = picture[n];
					n++;
				}
		break;

		case 5:
			for(y = y1; y <= y2; y++)
				for(x = x1; x <= x2; x++){
					if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
						if(vmem[(y << 8) + (y << 6) + x] == 0)
							vmem[(y << 8) + (y << 6) + x] = picture[n];
					n++;
				}
		break;
	}
}

void scaleput(int x1, int y1, float xscale, float yscale, unsigned char far *pic, char mode){
	int x, y, px, py, my, cy;
	const int xsize = (float)pic[0] / xscale;
	const int ysize = (float)pic[1] / yscale;
	const int ixscale = xscale * 256;
	const int iyscale = yscale * 256;

	for(y = 0; y <= ysize; y++){
		py = y * iyscale >> 8;
		my = py * pic[0];
		cy = y + y1;
		for(x = 0; x <= xsize; x++){
			px = x * ixscale >> 8;
			putpixel(x + x1, cy, pic[px + my + 2], mode);
		}
	}
}

void calcrotarea(int cx, int cy, int xoffset, int yoffset, float ang, float scale, unsigned char far *pic, int *x1, int *y1, int *x2, int *y2){
	int px[4], py[4], minx, miny, maxx, maxy;
	int dx, dy, n, tmp;
	const int SINE = sin(ang) * 256;
	const int COSINE = cos(ang) * 256;

	scale += !scale;   /**** prevents division by zero ******/

	px[2] = px[0] = -xoffset;
	py[1] = py[0] = -yoffset;
	px[1] = px[3] = px[0] + pic[0] - 1;
	py[2] = py[3] = py[0] + pic[1] - 1;

	for(n = 0; n < 4; n++){
		tmp = ((long)px[n] * COSINE - (long)py[n] * SINE) >> 8;
		py[n] = ((long)px[n] * SINE + (long)py[n] * COSINE) >> 8;
		px[n] = tmp;
		px[n] *= scale;
		py[n] *= scale;
		px[n] += cx;
		py[n] += cy;
	}

	*x1 = *x2 = px[0];
	*y1 = *y2 = py[0];
	for(n = 1; n < 4; n++){
		if(px[n] < *x1) *x1 = px[n];
		if(py[n] < *y1) *y1 = py[n];
		if(px[n] > *x2) *x2 = px[n];
		if(py[n] > *y2) *y2 = py[n];
	}
}

void rotput(int centerx, int centery, int xoffset, int yoffset, float ang, float scale, unsigned char far *pic, char mode){
/***
	As !!UUUUUUGGGGGLLLLLLYYYYYYY!! as this is with the case structure around
	the for loops - rather than vice versa - it's much faster, which is an
	absolute necessity.

	The chunks of code defined here at the top of the function are defined
	so as to make the case structure more ledgible, as these are both
	used in each case.

	Please note: The braces around each asm statement in the second definition
	are there because without them, asm requires a newline character, which
	I can't add as part of the #define.
***/


#define HANDLE_MATH_IN_X_LOOP 					\
	deltax = x - centerx;						\
	rotx = deltax * COSINE2 - deltaymulSINE2;	\
	roty = deltax * SINE2 + deltaymulCOSINE2;	\
	rotx >>= 6;									\
	roty >>= 6;									\
	rotx += xoffset;							\
	roty += yoffset;

#define HANDLE_MATH_IN_Y_LOOP 			\
	asm{mov ax, y}						\
	asm{mov bx, centery}				\
	asm{sub ax, bx}						\
	asm{mov deltay, ax}					\
	deltaymulSINE2 = deltay * SINE2;	\
	deltaymulCOSINE2 = deltay * COSINE2;\
	ymul320 = (y << 8) + (y << 6);

	const int xsize = pic[0], ysize = pic[1];
	int px[4], py[4], minx, miny, maxx, maxy;
	int deltax, deltay, n, tmp, rotx, roty, x, y;
	const int SINE = sin(ang) * 64;
	const int COSINE = cos(ang) * 64;
	const int SINE2 = sin(-ang) * 64 / (scale + !scale);
	const int COSINE2 = cos(-ang) * 64 / (scale + !scale);
	int deltaymulSINE2, deltaymulCOSINE2, ymul320;

	scale += !scale;   /**** prevents division by zero ******/

	px[2] = px[0] = -xoffset;
	py[1] = py[0] = -yoffset;
	px[1] = px[3] = px[0] + xsize - 1;
	py[2] = py[3] = py[0] + ysize - 1;

	for(n = 0; n < 4; n++){
		tmp = ((long)px[n] * COSINE - (long)py[n] * SINE) >> 6;
		py[n] = ((long)px[n] * SINE + (long)py[n] * COSINE) >> 6;
		px[n] = tmp;
		px[n] *= scale;
		py[n] *= scale;
		px[n] += centerx;
		py[n] += centery;

		if(px[n] < minx || !n) minx = px[n];
		if(py[n] < miny || !n) miny = py[n];
		if(px[n] > maxx || !n) maxx = px[n];
		if(py[n] > maxy || !n) maxy = py[n];
	}

	if(minx < 0) minx = 0;		if(maxx < 0) maxx = 0;
	if(miny < 0) miny = 0;		if(maxy < 0) maxy = 0;
	if(minx > 319) minx = 319;	if(maxx > 319) maxx = 319;
	if(miny > 199) miny = 199;	if(maxy > 199) maxy = 199;

	maxx++; /* The ++'s here let me use a < comaprison in the for loops, */
	maxy++; /* rather than a <=, which is a tad slower. */

	switch(mode){
		case 0:
			for(y = miny; y < maxy; y++){
				HANDLE_MATH_IN_Y_LOOP
				for(x = minx; x < maxx; x++){
					HANDLE_MATH_IN_X_LOOP
					if(rotx >= 0 && rotx < xsize && roty >= 0 && roty < ysize){
						vmem[x + ymul320] = pic[2 + rotx + xsize * roty];
					}
				}
			}
			break;

		case 1:
			for(y = miny; y < maxy; y++){
				HANDLE_MATH_IN_Y_LOOP
				for(x = minx; x < maxx; x++){
					HANDLE_MATH_IN_X_LOOP
					if(rotx >= 0 && rotx < xsize && roty >= 0 && roty < ysize){
						vmem[x + ymul320] ^= pic[2 + rotx + xsize * roty];
					}
				}
			}
			break;

		case 2:
			for(y = miny; y < maxy; y++){
				HANDLE_MATH_IN_Y_LOOP
				for(x = minx; x < maxx; x++){
					HANDLE_MATH_IN_X_LOOP
					if(rotx >= 0 && rotx < xsize && roty >= 0 && roty < ysize){
						vmem[x + ymul320] |= pic[2 + rotx + xsize * roty];
					}
				}
			}
			break;

		case 3:
			for(y = miny; y < maxy; y++){
				HANDLE_MATH_IN_Y_LOOP
				for(x = minx; x < maxx; x++){
					HANDLE_MATH_IN_X_LOOP
					if(rotx >= 0 && rotx < xsize && roty >= 0 && roty < ysize){
						vmem[x + ymul320] &= pic[2 + rotx + xsize * roty];
					}
				}
			}
			break;

		case 4:
			for(y = miny; y < maxy; y++){
				HANDLE_MATH_IN_Y_LOOP
				for(x = minx; x < maxx; x++){
					HANDLE_MATH_IN_X_LOOP
					if(rotx >= 0 && rotx < xsize && roty >= 0 && roty < ysize){
						if(pic[2 + rotx + xsize * roty])
							vmem[x + ymul320] = pic[2 + rotx + xsize * roty];
					}
				}
			}
			break;

		case 5:
			for(y = miny; y < maxy; y++){
				HANDLE_MATH_IN_Y_LOOP
				for(x = minx; x < maxx; x++){
					HANDLE_MATH_IN_X_LOOP
					if(rotx >= 0 && rotx < xsize && roty >= 0 && roty < ysize){
						if(!vmem[x + ymul320])
							vmem[x + ymul320] = pic[2 + rotx + xsize * roty];
					}
				}
			}
			break;

		case 6:
			for(y = miny; y < maxy; y++){
				HANDLE_MATH_IN_Y_LOOP
				for(x = minx; x < maxx; x++){
					HANDLE_MATH_IN_X_LOOP
					if(rotx >= 0 && rotx < xsize && roty >= 0 && roty < ysize){
						vmem[x + ymul320] += pic[2 + rotx + xsize * roty];
					}
				}
			}

	}
#undef HANDLE_MATH_IN_Y_LOOP
#undef HANDLE_MATH_IN_X_LOOP
}

void roterase(int x1, int y1, int x2, int y2, int centerx, int centery, float ang, float scale, unsigned char far *pic){
	const int SINE = sin(ang) * 64 / scale;
	const int COSINE = cos(ang) * 64 / scale;
	const int xsize = pic[0], ysize = pic[1];
	int deltax, deltay, rotx, roty, junk;
	int deltaymulSINE, deltaymulCOSINE, jumpsize;
	int leftdeltacenter, rightdeltacenter, bottomdeltacenter;
	unsigned char far *pptr;

	if(x1 < 0) x1 = 0; if(x1 > 319) x1 = 319;
	if(x2 < 0) x2 = 0; if(x2 > 319) x2 = 319;
	if(y1 < 0) y1 = 0; if(y1 > 199) y1 = 199;
	if(y2 < 0) y2 = 0; if(y2 > 199) y2 = 199;

	/****
		The ++'s are here so that I can use a < comparison in the for loops,
		rather than	a <=, which takes more processor time.
	****/
	x2++;
	y2++;

	/*
	pptr is the pointer aimed at the computer screen. I'm not too sure if
	it's any faster to increment pptr than it is to refer to vmem[n], but
	I think it is, and my testing didn't show any slowing down (although it
	didn't show any speeding up either).
	*/
	pptr = vmem + (y1 << 8) + (y1 << 6) + x1;

	/* jumpsize is the amount pptr is increased by to go from the end of
	   one line to the beginning of the of the next.  (rather than using
	   multiplication to calculate it)
	*/
	jumpsize = 320 - (x2 - x1);

	leftdeltacenter = x1 - centerx;
	rightdeltacenter = x2 - centerx;
	bottomdeltacenter = y2 - centery;
	for(deltay = y1 - centery; deltay < bottomdeltacenter; deltay++){

		deltaymulSINE = deltay * SINE;
		deltaymulCOSINE = deltay * COSINE;

		/***
			The reason why I have a leftdeltacenter in the x loop, and not a
			topdeltacenter in the y loop is this: The calculation of
			leftdeltacenter would be done y2 - y1 times if it was not assigned
			if a variable outside the loop.  On the other hand, topdeltacenter
			only gets calculated once, so assigning it to a variable would be
			a waste of time and memory.
		***/

		for(deltax = leftdeltacenter; deltax < rightdeltacenter; deltax++){
			rotx = deltax * COSINE - deltaymulSINE;
			roty = deltax * SINE + deltaymulCOSINE;

			rotx >>= 6;
			roty >>= 6;

			/****** add the rotated position to the center point ******/
			rotx += centerx;
			roty += centery;

			/****** modularize rotx & roty to fit the image ******/
			while(rotx < 0) rotx += xsize;
			while(roty < 0) roty += ysize;
			rotx %= xsize;
			roty %= ysize;

			*pptr = pic[2 + rotx + xsize * roty];
			pptr++;
		}

		pptr += jumpsize;
	}
}

void erase(int x1, int y1, int x2, int y2, int xoffset, int yoffset, unsigned char far *pic){
	int x, y, cx, cy, ymul320, xpicsize = pic[0], ypicsize = pic[1];

	if(x1 > x2) { x1 ^= x2; x2 ^= x1; x1 ^= x2; }
	if(y1 > y2) { y1 ^= y2; y2 ^= y1; y1 ^= y2; }

	if(y1 < 0) y1 = 0;
	if(y2 < 0) y2 = 0;
	if(y1 > 199) y1 = 199;
	if(y2 > 199) y2 = 199;
	if(x1 < 0) x1 = 0;
	if(x2 < 0) x2 = 0;
	if(x1 > 319) x1 = 319;
	if(x2 > 319) x2 = 319;


	for(y = y1; y <= y2; y++){
		cy = (y + yoffset) % ypicsize;
		ymul320 = (y << 8) + (y << 6);

		for(x = x1; x <= x2; x++){
			cx = (x + xoffset) % xpicsize;
			vmem[ymul320 + x] = pic[cx + xpicsize * cy + 2];
		}
	}
}

char touching(int x1, int y1, char *pic1, int x2, int y2, char *pic2, char mode){
	int x, y;
	char returnval = 0;

	if((x1 >= x2 - pic1[0]) &&
	  (x1 <= x2 + pic2[0]) &&
	  (y1 >= y2 - pic1[1]) &&
	  (y1 <= y2 + pic2[1]))
	{
		if(!mode) returnval = 1;
		else{
			for(x = x1; x < x1 + pic1[0]; x++){
				for(y = y1; y < y1 + pic1[1]; y++){
					if((x >= x2) && (x <= x2 + pic2[0] - 1) &&
					  (y >= y2) && (y <= y2 + pic2[1] - 1))
					{
						if(pic1[2 + x - x1 + pic1[0] * (y - y1)] &&
						   pic2[2 + x - x2 + pic2[0] * (y - y2)])
							returnval = 1;
					}
				}
			}
		}

	}

	return returnval;
}

/****************************** File I/O ******************************/
void writebit(FILE *fout, char bitval){
	static unsigned char bit[8], numbits = 0;
	unsigned char byte, n;

	if(bitval == -1){ /* force write despite numbits < 8 */
		for(;numbits < 8; numbits++)
			bit[numbits] = 0;
	}else{
		bit[numbits] = (bitval != 0);
		numbits++;
	}

	if(numbits >= 8){
		byte = 0;
		for(n = 0; n < 8; n++){
			if(bit[n])
				byte += 1 << (7 - n);
		}
		fputc(byte, fout);
		numbits = 0;
	}
}

char readbit(FILE *fin){
	char returnval = 0;
	static unsigned char byte, bitsleft = 0;


	if(!feof(fin) || bitsleft){
		if(bitsleft == 0){
			byte = fgetc(fin);
			bitsleft = 8;
		}

		bitsleft--;
		if(byte >=  1 << bitsleft){
			byte -= (1 << bitsleft);
			returnval = 1;
		}
		return returnval;
	}else{
		return -1;
	}
}

void save_font(unsigned char *name){
	int c, n;
	FILE *fout;

	if(font){
		fout = fopen(name, "wb");

		fputc(xfontsize, fout);
		fputc(yfontsize, fout);

		for(c = 0; c < 256; c++){
			for(n = 2; n < (xfontsize * yfontsize + 2); n++){
				//fputc(font[c][n], fout);
				writebit(fout, font[c][n] != 0);
			}
		}
		writebit(fout, -1);

		fclose(fout);
	}
}

void unload_font(void){
	int n;
	if(font){
		for(n = 0; n < 256; n++)
			if(font[n])
				free(font[n]);
		free(font);
		font = NULL;
	}
}

char load_font(unsigned char *name){
	int c, n, bad = 0;
	FILE *fname;
	fname = fopen(name, "rb");

	xfontsize = fgetc(fname);
	yfontsize = fgetc(fname);

	/******* free previously loaded font *******/
	if(font != NULL) unload_font();

	/******** allocate memory for new font *******/
	font = (unsigned char **)malloc(256 * sizeof(char *));
	if(font){
		for(n = 0; n < 256; n++){
			font[n] = (char *)malloc(xfontsize * yfontsize + 2);
			if(!font[n]) bad = 1;
		}
		if(bad){
			for(n = 0; n < 256; n++)
				if(font[n]) free(font[n]);
			free(font);
			font = NULL;
		}
	}else bad = 1;

	/******** load new font **********/
	for(c = 0; (c < 256) && !bad; c++){
		font[c][0] = xfontsize;
		font[c][1] = yfontsize;
		for(n = 2; n < (xfontsize * yfontsize + 2); n++){
			font[c][n] = readbit(fname);
		}
	}
	fclose(fname);

	return (!bad);
}

void savepic(char far *image, int arraysize, unsigned char *filename){
	int n;
	FILE *picfile;
	picfile = fopen(filename, "wb");
	n = image[0] * image[1] + 2;
	if(arraysize < n) n = arraysize;
	fwrite((void *)image, 1, n, picfile);
	fclose(picfile);
}

void loadpic(char far *image, int arraysize, unsigned char *filename){
	int n, size;
	FILE *picfile;
	picfile = fopen(filename, "rb");

	image[0] = fgetc(picfile);
	image[1] = fgetc(picfile);
	size = image[0] * image[1] + 2;
	if(size > arraysize) size = arraysize;
	size -= 2;
	fread((void *)&image[2], 1, size, picfile);
	fclose(picfile);
}

/**************************** user I/O *******************************/
unsigned char choose(unsigned char *title, char numbuttons, ...){
	typedef struct{
		char *text, character;
		int x1, x2, y1, y2;
	}buttontype;

	unsigned char memgood, *underbox, fini = 0;
	unsigned char keypressed, buttonused;
	buttontype *button;
	va_list arg_list;
	int buttonswidth, titlewidth;
	int n, boxwidth, boxx1, boxx2, boxy1, boxy2, x, y;

	/***** allocate enough memory to store all of the buttons *****/
	n = (4 * sizeof(int) + sizeof(char *) + sizeof(char)) * numbuttons;
	button = (buttontype *)malloc(n);
	memgood = ((button != NULL) || (n == 0));

	/***** read the various arguments *****/
	va_start(arg_list, numbuttons);
	for(n = 0; n < numbuttons; n++){
		button[n].text = strdup(va_arg(arg_list, char *));
		button[n].character = va_arg(arg_list, char);
		if(button[n].text == NULL) memgood = 0;
	}
	va_end(arg_list);

	/***** calculate collective width of buttons *****/
	if(memgood){
		buttonswidth = xfontsize;
		for(n = 0; n < numbuttons; n++){
			buttonswidth += (strlen(button[n].text) + 1) * xfontsize;
		}
	}
	boxwidth = buttonswidth;

	titlewidth = (strlen(title) + 2) * xfontsize;
	if(boxwidth < titlewidth) boxwidth = titlewidth;

	/***** calculate size and location of box *****/
	boxx1 = 160 - (boxwidth / 2);
	boxx2 = boxx1 + boxwidth - 1;
	boxy1 = 100 - 2 * yfontsize;
	boxy2 = boxy1 + 4 * yfontsize - 1;

	underbox = (char *)malloc((boxx2 - boxx1 + 1) * (boxy2 - boxy1 + 1) + 2);
	memgood *= (underbox != NULL);

	if(memgood){
		/***** get the image beneath the box *****/
		//showmouse(0);
		get(boxx1, boxy1, boxx2, boxy2, underbox);

		/***** draw the box *****/
		box(boxx1, boxy1, boxx2 - 1, boxy2 - 1, 22, 4);
		box(boxx1 + 1, boxy1 + 1, boxx2, boxy2, 18, 4);
		box(boxx1 + 1, boxy1 + 1, boxx2 - 1, boxy2 - 1, 20, 0);

		/***** determine the locations of, and draw, the buttons *****/
		x = 160 - buttonswidth / 2;
		for(n = 0; n < numbuttons; n++){
			x += xfontsize;
			button[n].x1 = x - 2;
			x += (xfontsize * strlen(button[n].text));
			button[n].x2 = x + 2;
			button[n].y2 = boxy2 - (yfontsize / 2) + 1;
			button[n].y1 = button[n].y2 - yfontsize - 2;
			box(button[n].x1 + 1, button[n].y1 + 1, button[n].x2, button[n].y2, 18, 4);
			box(button[n].x1, button[n].y1, button[n].x2 - 1, button[n].y2 - 1, 22, 4);
			box(button[n].x1 + 1, button[n].y1 + 1, button[n].x2 - 1, button[n].y2 - 1, 20, 0);
			drawtext(button[n].x1 + 2, button[n].y1 + 1, button[n].text, 15, 4);
		}

		/***** draw the title *****/
		drawtext(160 - titlewidth / 2 + xfontsize, boxy1 + yfontsize / 2, title, 15, 4);

		showmouse(1);
		while(!fini){
			/***** handle keyboard IO *****/
			if(kbhit()){

				keypressed = getch();

				/***** test for hotkeys *****/
				for(n = 0; n < numbuttons && !fini; n++){
					if(keypressed == button[n].character){
						fini = 1;
						buttonused = n;
					}
				}
			}

			/***** handle mouse IO *****/
			if(buttonpressed()){
				x = mousex();
				y = mousey();
				for(n = 0; n < numbuttons; n++){
					if((button[n].x1 <= x) && (button[n].x2 >= x) &&
					   (button[n].y1 <= y) && (button[n].y2 >= y))
						break;
				}
				if(n < numbuttons){
					fini = 1;
					buttonused = n;
				}
			}
		}
		showmouse(0);
		put(boxx1, boxy1, underbox, 0);
		//showmouse(1);
	}

	for(n = 0; n < numbuttons; n++)
		if(button[n].text) free(button[n].text);
	if(button) free(button);
	if(underbox) free(underbox);

	return buttonused;
}

unsigned char *taketext(unsigned char *title, char *buttonused, char numbuttons, ...){
	typedef struct{
		char *text, character;
		int x1, x2, y1, y2;
	}buttontype;

	const int maxstringlength = 1024;
	unsigned char returnval[1024], memgood, *underbox, fini = 0;
	unsigned char keypressed, *drawstring, insertmode = 1;
	buttontype *button;
	va_list arg_list;
	int buttonswidth, titlewidth;
	int n, boxwidth, boxx1, boxx2, boxy1, boxy2, x, y;
	int textx1, textx2, texty, displaywidth, stringlength;
	int coffset, doffset, cx;

	/***** allocate enough memory to store all of the buttons *****/
	n = (4 * sizeof(int) + sizeof(char *) + sizeof(char)) * numbuttons;
	button = (buttontype *)malloc(n);
	memgood = ((button != NULL) || (n == 0));

	/***** read the various arguments *****/
	va_start(arg_list, numbuttons);
	for(n = 0; n < numbuttons; n++){
		button[n].text = strdup(va_arg(arg_list, char *));
		button[n].character = va_arg(arg_list, char);
		if(button[n].text == NULL) memgood = 0;
	}
	va_end(arg_list);

	/***** calculate collective width of buttons *****/
	if(memgood){
		buttonswidth = xfontsize;
		for(n = 0; n < numbuttons; n++){
			buttonswidth += (strlen(button[n].text) + 1) * xfontsize;
		}
	}
	boxwidth = buttonswidth;

	titlewidth = (strlen(title) + 2) * xfontsize;
	if(boxwidth < titlewidth) boxwidth = titlewidth;

	/***** calculate size and location of box *****/
	boxx1 = 160 - (boxwidth / 2);
	boxx2 = boxx1 + boxwidth - 1;
	if(numbuttons){
		boxy1 = 100 - 3 * yfontsize;
		boxy2 = boxy1 + 6 * yfontsize - 1;
	}else{
		boxy1 = 100 - 2 * yfontsize;
		boxy2 = boxy1 + 4 * yfontsize - 1;
	}

	underbox = (char *)malloc((boxx2 - boxx1 + 1) * (boxy2 - boxy1 + 1) + 2);
	memgood *= (underbox != NULL);

	if(memgood){
		/***** get the image beneath the box *****/
		get(boxx1, boxy1, boxx2, boxy2, underbox);

		/***** draw the box *****/
		box(boxx1, boxy1, boxx2 - 1, boxy2 - 1, 22, 4);
		box(boxx1 + 1, boxy1 + 1, boxx2, boxy2, 18, 4);
		box(boxx1 + 1, boxy1 + 1, boxx2 - 1, boxy2 - 1, 20, 0);

		/***** determine the locations of, and draw, the buttons *****/
		x = 160 - buttonswidth / 2;
		for(n = 0; n < numbuttons; n++){
			x += xfontsize;
			button[n].x1 = x - 2;
			x += (xfontsize * strlen(button[n].text));
			button[n].x2 = x + 2;
			button[n].y2 = boxy2 - (yfontsize / 2) + 1;
			button[n].y1 = button[n].y2 - yfontsize - 2;
			box(button[n].x1 + 1, button[n].y1 + 1, button[n].x2, button[n].y2, 18, 4);
			box(button[n].x1, button[n].y1, button[n].x2 - 1, button[n].y2 - 1, 22, 4);
			box(button[n].x1 + 1, button[n].y1 + 1, button[n].x2 - 1, button[n].y2 - 1, 20, 0);
			drawtext(button[n].x1 + 2, button[n].y1 + 1, button[n].text, 15, 4);
		}

		/***** draw the title *****/
		drawtext(160 - titlewidth / 2 + xfontsize, boxy1 + yfontsize / 2, title, 15, 4);

		/***** draw the text box *****/
		n = boxx2 - boxx1 - 2 * xfontsize;
		n -= (n % xfontsize);
		textx1 = 160 - n / 2;
		textx2 = textx1 + n;
		displaywidth = n / xfontsize;
		drawstring = (char *)malloc(displaywidth + 1);
		memgood = drawstring != NULL;

		if(numbuttons)
			texty = button[0].y1 - 1.5 * yfontsize;
		else
			texty = boxy2 - 1.5 * yfontsize;
		box(textx1 - 2, texty - 2, textx2 + 1, texty + yfontsize + 1, 18, 0);
		box(textx1 - 1, texty - 1, textx2 + 2, texty + yfontsize + 2, 22, 0);
		box(textx1 - 1, texty - 1, textx2 + 1, texty + yfontsize + 1, 0, 0);

		showmouse(1);
		memset(returnval, NULL, maxstringlength);
		coffset = doffset = stringlength = 0;
		while(!fini && drawstring){
			/***** handle keyboard IO *****/
			if(kbhit()){
				/***** hide the cursor *****/
				box(textx1 + xfontsize * (stringlength + coffset - doffset), texty,
					textx1 + xfontsize * (stringlength + coffset - doffset) + (xfontsize - 2) * (1 - insertmode), texty + yfontsize - 1,
					0, 0);

				keypressed = getch();

				/***** test for hotkeys *****/
				for(n = 0; n < numbuttons && !fini; n++){
					if(keypressed == button[n].character){
						fini = 1;
						*buttonused = n;
					}
				}
				/***** test for other keys *****/
				if(!fini){
					switch(keypressed){
						case 0:
							switch(getch()){
								case 'G': //home
									coffset = -stringlength;
								break;
								case 'O':
									coffset = 0;
								break;
								case 'K': //left
									if(- coffset < stringlength)
										coffset--;
								break;
								case 'M': //right
									if(coffset < 0)
										coffset++;
								break;
								case 83: //del
									if(- coffset > 0){
										for(n = coffset; n < 0; n++)
											returnval[stringlength + n] = returnval[stringlength + n + 1];
										stringlength--;
										coffset++;
										returnval[stringlength] = NULL;
									}
								break;
								case 82: //insert
									insertmode = 1 - insertmode;
								break;
							}
						break;
						case 8: //backspace
							if(- coffset < stringlength){
								for(n = coffset; n < 0; n++)
									returnval[stringlength + n - 1] = returnval[stringlength + n];
								stringlength--;
								returnval[stringlength] = NULL;
							}
						break;
						case 13: //enter
							*buttonused = 0;
							fini = 1;
						break;
						default: //alphanumagrammatic characters
							if((keypressed >= 32) && (keypressed <= 126) && (stringlength < maxstringlength)){
								if(insertmode)
									for(n = 0; n > coffset; n--)
										returnval[stringlength + n] = returnval[stringlength + n - 1];

								returnval[stringlength + coffset] = keypressed;

								if(insertmode || coffset == 0)
									stringlength++;
								else
									coffset++;
							}
					}
				}

				/***** redraw the string *****/
				n = stringlength + coffset;
				if(n > doffset + displaywidth - !insertmode)
					doffset = n - displaywidth + !insertmode;
				if(n < doffset)
					doffset = n;

				memset(drawstring, 32, displaywidth + 1);
				drawstring[displaywidth] = NULL;
				strncpy(drawstring, &returnval[doffset], displaywidth);
				for(n = 0; n < displaywidth; n++)
					if(drawstring[n] == NULL) drawstring[n] = 32;

				showmouse(0);
				drawtext(textx1, texty, drawstring, 15, 0);

				/***** draw the cursor *****/
				box(textx1 + xfontsize * (stringlength + coffset - doffset), texty,
					textx1 + xfontsize * (stringlength + coffset - doffset) + (xfontsize - 2) * (1 - insertmode), texty + yfontsize - 1,
					14, 1);
				showmouse(1);
			}

			/***** handle mouse IO *****/
			if(buttonpressed()){
				x = mousex();
				y = mousey();
				for(n = 0; n < numbuttons; n++){
					if((button[n].x1 <= x) && (button[n].x2 >= x) &&
					   (button[n].y1 <= y) && (button[n].y2 >= y))
						break;
				}
				if(n < numbuttons){
					fini = 1;
					*buttonused = n;
				}
			}
		}
		showmouse(0);

		put(boxx1, boxy1, underbox, 0);
	}

	for(n = 0; n < numbuttons; n++)
		if(button[n].text) free(button[n].text);
	if(button) free(button);
	if(underbox) free(underbox);
	if(drawstring) free(drawstring);

	return returnval;
}

void adjust_value(char *title, int *value, int min, int max){
	unsigned char *underbox;
	char junk[5], fini = 0, val = *value;
	int bx1, by1, bx2, by2, tx, ty, nx1, ny1, nx2, ny2, mdx = 0, mdy = 0;

	tx = 160 - (xfontsize * strlen(title)) / 2;
	ty = 100 - 2 * yfontsize;
	bx1 = tx - 10;
	bx2 = tx + (xfontsize * strlen(title)) + 10;
	by1 = ty - 2;
	by2 = by1 + 4 * yfontsize;
	nx1 = (bx1 + bx2) / 2 - 2 * xfontsize;
	nx2 = nx1 + 4 * xfontsize;
	ny1 = ty + yfontsize + 4;
	ny2 = ny1 + yfontsize;

	if(max < min){	/* toggle values of max and min if min is greater */
		max ^= min;
		min ^= max;
		max ^= min;
	}

	showmouse(0);
	if((underbox = (char *)malloc((bx2 - bx1 + 1) * (by2 - by1 + 1) + 1)) != NULL){
		get(bx1, by1, bx2, by2, underbox);
		box(bx1 + 1, by1 + 1, bx2, by2, 18, 4);
		box(bx1, by1, bx2 - 1, by2 - 1, 22, 4);
		box(bx1 + 1, by1 + 1, bx2 - 1, by2 - 1, 20, 0);

		drawtext(tx, ty, title, 15, 4);

		box(nx1 - 3, ny1 - 3, nx2 + 2, ny2 + 2, 18, 4);
		box(nx1 - 2, ny1 - 2, nx2 + 3, ny2 + 3, 22, 4);
		box(nx1 - 2, ny1 - 2, nx2 + 2, ny2 + 2, 0, 0);
		itoa(val, junk, 10);
		drawtext(nx1, ny1, junk, 15, 0);

		while(buttonpressed());
		while(!fini){
			do{
				readmousediff();
				mdx += mousediff.deltax;
				mdy += mousediff.deltay;
				if(mdx > 10 || mdy < -10){
					mdx = mdy = 0;
					box(nx1 - 2, ny1 - 2, nx2 + 2, ny2 + 2, 0, 0);
					val = val + 1;
					if(val > max) val = min;
					itoa(val, junk, 10);
					drawtext(nx1, ny1, junk, 15, 0);
				}else if(mdx < - 10 || mdy > 10){
					mdx = mdy = 0;
					box(nx1 - 2, ny1 - 2, nx2 + 2, ny2 + 2, 0, 0);
					val = val - 1;
					if(val < min) val = max;
				itoa(val, junk, 10);
					drawtext(nx1, ny1, junk, 15, 0);
				}
				if(buttonpressed()){
					*value = val;
					fini = 1;
				}
			}while(buttonpressed());
			if(kbhit()){
				switch(getch()){
					case 27: fini = 1; break;
					case 13: case 32:
						fini = 1;
						*value = val;
					break;
					case 0:
						switch(getch()){
							case 'H': case 'M':
								box(nx1 - 2, ny1 - 2, nx2 + 2, ny2 + 2, 0, 0);
								val = (val + 1);
								if(val > max) val = min;
								itoa(val, junk, 10);
								drawtext(nx1, ny1, junk, 15, 0);
							break;
							case 'K': case 'P':
								box(nx1 - 2, ny1 - 2, nx2 + 2, ny2 + 2, 0, 0);
								val = val - 1;
								if(val < min) val = max;
								itoa(val, junk, 10);
								drawtext(nx1, ny1, junk, 15, 0);
							break;
						}
					break;
				}
			}
		}
		put(bx1, by1, underbox, 0);
		free(underbox);
	}
	showmouse(1);
}

